"""
This file defines the CLI options in the front subcommand.
"""
import os
from pathlib import Path
from click import option, command
from pydantic import BaseModel
from gpttui.database.base import AbstractDB, DatabasesEnum
from gpttui.database.sqlite import SqliteDB
from gpttui.models.base import AbstractModel, ModelsEnum
from gpttui.models.chatsonic import ChatSonicConf, ChatSonicModel
from gpttui.models.openai import OpenAIModel, OpenAIConf
from gpttui.tui.app import GptApp
from gpttui.tui.config import config_file
from typing import Dict, Type

DBS: Dict[DatabasesEnum, Type[AbstractDB]] = {
        DatabasesEnum.SQLITE: SqliteDB
        }
MODELS: Dict[ModelsEnum, Type[AbstractModel]] = {
        ModelsEnum.OPENAI: OpenAIModel,
        ModelsEnum.CHATSONIC: ChatSonicModel
        }
CONFS: Dict[ModelsEnum, Type[BaseModel]] = {
        ModelsEnum.OPENAI: OpenAIConf,
        ModelsEnum.CHATSONIC: ChatSonicConf
        }

@command()
@option(
    "--database_kind",
    type=DatabasesEnum,
    default=DatabasesEnum.SQLITE,
    help="Database to store the messages."
    )
@option(
    "--database_name",
    type=str,
    default=Path(os.environ["HOME"]) / ".config/gpttui/database.sqlite",
    help="Connection string for the database."
    )
@option(
    "--session",
    type=str,
    default="default_session",
    help="Session (chat) to use."
    )
@option(
    "--model_kind",
    type=ModelsEnum,
    default=ModelsEnum.OPENAI,
    help="Which model to use."
    )
@option(
    "--context",
    type=str,
    default="You are an AI assistant",
    help="Context for the model."
    )
@option(
    "--config",
    type=Path,
    default=Path(os.environ["HOME"]) / ".config/gpttui/openai.json",
    help="Context for the model."
    )
def front(
    database_kind: DatabasesEnum,
    database_name: str,
    session: str,
    model_kind: ModelsEnum,
    context: str,
    config: Path
    ) -> None:
    """
    Determines what to do when the front subcommand is launched.

    Parameters
    ----------
    database_kind : DatabasesEnum
        Which database to use.
    database_name : str
        Connection string to the database.
    session : str
        Session name.
    model_kind : ModelsEnum
        Which model to use.
    model_name : str
        Model name.
    context : str
        Context for the model.
    config : Path
        Configuration file.
    """
    db = (
            DBS[database_kind]()
            .setup(database=database_name)
            )
    db.create_session(session_name=session)
    cfg = config_file(config, CONFS[model_kind])

    model = (
            MODELS[model_kind]()
            .add_context(context=context)
            .setup(config=cfg, database=db, session_name=session)
            )
    app = (
            GptApp()
            .setup(model=model)
            )
    app.run()



"""
This file defines the main CLI.
"""
from click import group
from gpttui.tui.front import front
from gpttui.tui.config import css_config

@group()
def cli() -> None:
    css_config()

cli.add_command(front)



"""
This file defines the main TUI App.
"""
import os, re
import pyperclip
from pathlib import Path
from enum import Enum, auto
from typing import Any
from textual.app import App, ComposeResult
from textual.widgets import Input, Markdown, Static
from textual.containers import Container
from textual.events import Key
from gpttui.models.base import AbstractModel
from gpttui.tui.config import KeyBindings, keybindings_config

class ModeEnum(Enum):
    """
    This enum represents the possible modes of the application.
    """
    INSERT = auto()
    NORMAL = auto()

class NormalIndicator(Static):
    """
    Static container for the normal mode.
    """
    ...

class InsertIndicator(Static):
    """
    Static container for the insert mode.
    """
    ...

class UserInput(Input):
    """
    Represents the text input.
    """
    ...

class Prompt(Static):
    """
    The prompt contains the mode indicators and the text input.
    """
    def compose(self) -> ComposeResult:
        yield NormalIndicator("NORMAL", id="normal-indicator")
        yield InsertIndicator("INSERT", id="insert-indicator")
        yield Input(placeholder="Enter some text...")

class UserText(Static):
    """
    Static container to display the sender.
    """
    ...

class Message(Static):
    """
    A message contains the the sender and its text.

    Parameters
    ----------
    user : str
        Sender.
    message : str
        Text of the message.
    """

    def __init__(self, user:str, message: str, *args: Any, **kwargs: Any):
        super(Message, self).__init__(*args, **kwargs)
        self.user = user
        self.message = message

    def compose(self) -> ComposeResult:
        """
        Generator with the message components.

        Yields
        ------
        ComposeResult:
            Widgets in the message.

        """
        yield UserText(self.user)
        # BUG: seems like textual isn't able to render inline code compose markdowns.
        try:
            md = Markdown()
            md.update(self.message)
        except:
            md = Static(self.message)
        yield md

class Messages(Container):
    """
    A container that stores all messages.
    """

    def add_message(self, msg: str, user: str):
        """
        This method dynamically adds a message.

        Parameters
        ----------
        msg : str
            Message to display.
        user : str
            Sender of the message.
        """
        self.mount(Message(user=user, message=msg))
        self.scroll_end()

class GptApp(App):
    """
    This class defines the TUI App.

    Parameters
    ----------
    CSS_PATH : Path
        Path for the CSS styling file.
    KEYBINDINGS : KeyBindings
        Keybindings to use in the application.
    model : AbstractModel
        Model to use in the App.
    mode : ModeEnum
        Current mode in the App.
    chat_text : str
        Visible text that is show in the chat.
    normal_commands : str
        Mapping between a keybinding and its function in normal mode.
    insert_commands : str
        Mapping between a keybinding and its function in insert mode.
    """
    CSS_PATH : Path = Path(os.environ["HOME"]) / ".config/gpttui/style.css"
    KEYBINDINGS : KeyBindings = keybindings_config()
    model: AbstractModel

    def __init__(self, *args: Any, **kwargs: Any):
        super(GptApp, self).__init__(*args, **kwargs)
        self.mode = ModeEnum.NORMAL
        self.normal_commands = {
                self.KEYBINDINGS.insert: self.insert,
                self.KEYBINDINGS.quit: self.quit,
                self.KEYBINDINGS.yank: self.yank,
                self.KEYBINDINGS.paste: self.paste,
                self.KEYBINDINGS.clear: self.clear,
                self.KEYBINDINGS.delete: self.delete
                }
        self.insert_commands = {
                self.KEYBINDINGS.normal: self.normal,
                self.KEYBINDINGS.send: self.send,
                }

    def setup(self, model: AbstractModel) -> "GptApp":
        """
        Setups the App.

        Parameters
        ----------
        model : AbstractModel
            Specifies the model to use.

        Returns
        -------
        GptApp
            Instance of the app to use as a builder.
        """
        self.model = model
        return self

    def compose(self) -> ComposeResult:
        """
        Generator with the TUI components.

        Yields
        ------
        ComposeResult
            TUI components.
        """
        yield Prompt()
        yield Messages()

    async def on_key(self, event: Key) -> None:
        """
        Callback that is called when a key is pressed.

        Parameters
        ----------
        event : Key
            Event related to the key that was pressed.
        """
        if self.mode == ModeEnum.NORMAL:
            await self.handle_normal(event)
        elif self.mode == ModeEnum.INSERT:
            await self.handle_insert(event)

    async def handle_normal(self, event: Key) -> None:
        """
        Determines what to do in normal mode.

        Parameters
        ----------
        event : Key
            Event related to the key that was pressed.
        """
        f = self.normal_commands.get(event.key)
        if f is not None: await f()

    async def handle_insert(self, event: Key) -> None:
        """
        Determines what to do in insert mode.

        Parameters
        ----------
        event : Key
            Event related to the key that was pressed.
        """
        f = self.insert_commands.get(event.key)
        if f is not None: await f()

    async def insert(self):
        """
        Changes to insert mode.
        """
        self.add_class("insert-mode")
        self.query_one(Input).focus()
        self.mode = ModeEnum.INSERT

    async def clear(self):
        """
        Clears the historical messages.
        """
        self.query_one(Messages) #TODO

    async def yank(self):
        """
        Copies the last message into the clipboard.
        """
        msgs = self.model.database.get_messages(self.model.session_name)
        pyperclip.copy(msgs.values[-1].content)

    async def paste(self):
        """
        Pastes the clipboard into the prompt.
        """
        clipboard_text = pyperclip.paste()
        self.query_one(Input).insert_text_at_cursor(clipboard_text)

    async def normal(self):
        """
        Switch to normal mode.
        """
        self.remove_class("insert-mode")
        self.query_one(Input).reset_focus()
        self.mode = ModeEnum.NORMAL

    async def quit(self):
        """
        Exit the app.
        """
        self.model.database.close()
        self.exit()

    async def delete(self):
        """
        Deletes the prompt.
        """
        inp = self.query_one(Input)
        inp.action_delete_right_all()
        inp.action_delete_left_all()

    async def send(self):
        """
        Sends the text in the prompt to the model and shows the response.
        """
        inp = self.query_one(Input)
        text = inp.value
        messages = self.query_one(Messages)
        inp.action_delete_right_all()
        inp.action_delete_left_all()
        messages.add_message(msg=text, user="User")
        answer = await self.model.get_answer(text)
        messages.add_message(msg=answer, user="Assistant")



"""
Tests for the databases integration.
"""
import pytest, time
from gpttui.database.sqlite import SqliteDB
from gpttui.database.base import AbstractDB, Message, MessageWithTime

class TestSqliteDB:
    """
    Unittests for the sqlite database.
    """
    @staticmethod
    def setup_db() -> AbstractDB:
        """
        Initializes the database.

        Returns
        -------
        AbstractDB
            Initialized database.
        """
        db = SqliteDB().setup(database="test.db")
        return db

    @pytest.mark.parametrize("session_name", [f"chat{i}" for i in range(10)])
    def test_session(self, session_name: str):
        """
        Tests the session creation and deletion.

        Parameters
        ----------
        session_name : str
            Session name.
        """
        db = TestSqliteDB.setup_db()
        db.create_session(session_name)
        cursor = db.connection.cursor()
        cursor.execute("SELECT name FROM sqlite_schema;")
        *result, = map(lambda i: i[0], cursor.fetchall())
        assert session_name in result

        db.delete_session(session_name)

        cursor = db.connection.cursor()
        cursor.execute("SELECT name FROM sqlite_schema;")
        *result, = map(lambda i: i[0], cursor.fetchall())
        assert session_name not in result

    @pytest.mark.parametrize("message", ["hello", "testing", "hi"])
    def test_message(self, message: str):
        """
        Tests the save and retrieval of messages.

        Parameters
        ----------
        message : str
            Message to save.
        """
        db = TestSqliteDB.setup_db()
        db.create_session("test")
        msg = Message(role="user", content=message)
        msgt = MessageWithTime(message=msg, timestamp=int(time.time()))
        db.add_message(msgt, "test")

        messages = db.get_messages("test")
        msgt2 = messages.values[0].dict()
        assert msgt2 == msg.dict()
        db.delete_session("test")



"""
This module contains the integration with ChatSonic.
"""
import time
from typing import List
from pydantic import BaseModel
from gpttui.models.base import AbstractModel
from gpttui.database.base import Messages, Message, MessageWithTime, AbstractDB
try:
    import httpx
except ImportError:
    raise ImportError("Could not import chatsonic dependencies, please install it with:\n\tpip install gpttui[chatsonic]")

class ChatSonicConf(BaseModel):
    url : str = "https://api.writesonic.com/v2/business/content/chatsonic?engine=premium"
    api_key : str = ""
    enable_memory: bool = True
    enable_google_results: bool = True

class ChatSonicMessage(BaseModel):
    is_sent : bool
    message : str

class ChatSonicMessages(BaseModel):
    values : List[ChatSonicMessage]

class ChatSonicModel(AbstractModel):
    """
    This class allows loading and interacting with any openai model through its API.
    """
    config : ChatSonicConf

    def setup(self, config: ChatSonicConf, session_name: str, database: AbstractDB) -> "ChatSonicModel":
        """
        Initializes the model and collects credentials for OpenAI.

        Parameters
        ----------
        url : str
            Chatsonic URL.

        Returns
        -------
        OpenAIModel
            Instance of the model to be used as a builder.
        """
        self.config = config
        self.session_name = session_name
        self.database = database
        return self

    def last_messages(self) -> Messages:
        """
        Extracts the most recent messages from the database.

        Returns
        -------
        Messages
            Most recent messages.
        """
        self.database.create_session(session_name=self.session_name)
        last_msgs = self.database.get_messages(session_name=self.session_name)
        if not len(last_msgs.values):
            msg = MessageWithTime(
                    message=Message(role="system", content=self.context),
                    timestamp=int(time.time())
                    )
            self.database.add_message(msg=msg, session_name=self.session_name)
            return self.last_messages()
        return last_msgs

    @staticmethod
    def parse_messages(msgs: Messages) -> ChatSonicMessages:
        """
        This method parses general messages into chatsonic specific messages.

        Parameters
        ----------
        msgs : Messages
            Input messages.

        Returns
        -------
        ChatSonicMessages
            Parsed messages.
        """ 
        parsed_msgs = []
        for msg in msgs.values:
            if msg.role not in ["assistant", "user"]:
                continue
            is_sent = msg.role == "user"
            parsed_msgs.append(
                    ChatSonicMessage(is_sent=is_sent, message=msg.content)
                    )
        return ChatSonicMessages(values=parsed_msgs)

    async def get_answer(self, message: str) -> str:
        """
        Obtains an answer form any message.

        Parameters
        ----------
        message : str
            Input message.

        Returns
        -------
        str
            Generated response.
        """
        last_msgs = self.last_messages()
        new_msg = MessageWithTime(
                message=Message(role="user", content=message),
                timestamp=int(time.time())
                )
        self.database.add_message(msg=new_msg, session_name=self.session_name)
        last_msgs = self.last_messages()
        payload = {
                "enable_memory": self.config.enable_memory,
                "enable_google_results": self.config.enable_google_results,
                "input_text": self.context,
                "history_data": ChatSonicModel.parse_messages(last_msgs).dict()["values"]
                } 
        headers = {
                "accept": "application/json",
                "content-type": "application/json",
                "X-API-KEY": self.config.api_key
                }
        async with httpx.AsyncClient() as client:
            r = await client.post(self.config.url, json=payload, headers=headers)
            response = r.json()["message"]

        new_msg = MessageWithTime(
                message=Message(role="assistant", content=response),
                timestamp=int(time.time())
                )
        self.database.add_message(msg=new_msg, session_name=self.session_name)
        return response



"""
This module defines the general classes required to integrate different models.
"""
from abc import ABC, abstractmethod

from pydantic import BaseModel
from gpttui.database.base import AbstractDB
from enum import Enum

class ModelsEnum(Enum):
    """
    This enum defines the available models.
    """
    OPENAI = "OPENAI"
    CHATSONIC = "CHATSONIC"

class AbstractModel(ABC):
    """
    This abstract class defines any generative model.

    Attributes
    ----------
    model_name : str
        Model name.
    session_name : str
        Session name.
    context : str
        Context given to the model.
    database : AbstractDB
        Database to store the messages.
    """
    config: BaseModel
    session_name: str
    context: str
    database: AbstractDB

    def add_context(self, context: str) -> "AbstractModel":
        """
        Sets the context.

        Parameters
        ----------
        context : str
            Context given to the model.

        Returns
        -------
        AbstractModel
            Instance of the model to use as a builder.
        """
        self.context = context
        return self

    @abstractmethod
    def setup(self, config: BaseModel, session_name: str, database: AbstractDB) -> "AbstractModel":
        """
        This method is used to setup any model.

        Parameters
        ----------
        config : BaseModel
            Dataclass with the model's config options.
        session_name : str
            Session name.
        database : AbstractDB
            Database to store the messages.

        Returns
        -------
        AbstractModel
            Instance of the model to use as a builder.
        """
        ...

    @abstractmethod
    async def get_answer(self, message: str) -> str:
        """
        Generates an answer given an input message.

        Parameters
        ----------
        message : str
            Input text.

        Returns
        -------
        str
            Response.
        """
        ...



"""
This module defines the general classes that are required in the database module.
"""

from abc import ABC, abstractmethod
from typing import List, Any
from pydantic import BaseModel
from enum import Enum

class DatabasesEnum(Enum):
    """
    Enum that specifies the supported databases.
    """
    SQLITE = "SQLITE"

class Message(BaseModel):
    """
    Dataclass that contains a single message.

    Attributes
    ----------
    role : str
        Who wrote the message.
    content : str
        Message content.
    """
    role: str
    content: str

class Messages(BaseModel):
    """
    Dataclass that represents multiple messages.

    Attributes
    ----------
    values : List[Message]
        List of messages.
    """
    values: List[Message]

class MessageWithTime(BaseModel):
    """
    Dataclass that represents a message and its timestamp.

    Attributes
    ----------
    message : Message
        A message instance.
    timestamp : int
        Unix time.
    """
    message: Message
    timestamp: int

class AbstractDB(ABC):
    """
    Abstract class that represents any compatible database.

    Attributes
    ----------
    connection : Any
        Connection with any database.
    """
    connection: Any

    @abstractmethod
    def setup(self, **kwargs: str) -> "AbstractDB":
        """
        This method must be used to setup the database. For instance, to add connection strings, clients, among others.

        Parameters
        ----------
        kwargs : str
            Any string keyword argument.

        Returns
        -------
        AbstractDB
            Instance of the database to use the method as a builder.
        """
        ...

    @abstractmethod
    def create_session(self, session_name: str):
        """
        This method is intended to create a session.

        Parameters
        ----------
        session_name : str
            Name of the session.
        """
        ...

    @abstractmethod
    def delete_session(self, session_name: str):
        """
        This method is intended to delete a session.

        Parameters
        ----------
        session_name : str
            Name of the session.
        """
        ...

    @abstractmethod
    def add_message(self, msg: MessageWithTime, session_name: str):
        """
        Add a message to the database.

        Parameters
        ----------
        msg : MessageWithTime
            Message to store.
        session_name : str
            Name of the session.
        """
        ...

    @abstractmethod
    def get_messages(self, session_name: str) -> Messages:
        """
        Get the most recent messages from the database.

        Parameters
        ----------
        session_name : str
            Name of the session.
        """
        ...

    @abstractmethod
    def close(self):
        """
        This method allows closing the connection.
        """
        ...



"""
Defines the tests that are performed over models.
"""
import pytest
from gpttui.database.base import AbstractDB
from gpttui.models.base import AbstractModel
from gpttui.models.openai import OpenAIModel
from gpttui.database.sqlite import SqliteDB
from typing import Tuple

class TestOpenAi:
    """
    Tests that are used for any OpenAI model.
    """
    @staticmethod
    def setup_model() -> Tuple[AbstractDB, AbstractModel]:
        """
        Initialize the model and the database.

        Returns
        -------
        Tuple[AbstractDB, AbstractModel]
            Initialized database and model.
        """
        db = (
                SqliteDB()
                .setup(database="test.db")
                )
        model = (
                OpenAIModel()
                .add_context(context="You're an expert programmer")
                .setup(model_name="gpt-3.5-turbo", database=db, session_name="test")
                )
        return db, model

    @pytest.mark.parametrize("message", [
        "Is Python better than JS?",
        "How to write hello word in Python?"
        ])
    def test_generation(self, message: str):
        """
        Tests the generation from the models.

        Parameters
        ----------
        message : str
            Input message.
        """
        db, model = (
                TestOpenAi
                .setup_model()
                )
        db.create_session(session_name="test")
        answer = model.get_answer(message)
        db.delete_session(session_name="test")
        assert isinstance(answer, str)



"""
This module contains the integration with OpenAI models.
"""
try:
    import openai, time
    from openai.error import Timeout
except ImportError:
    raise ImportError("Could not import openai library, please install it with:\n\tpip install gpttui[openai]")
from pydantic import BaseModel
from gpttui.models.base import AbstractModel
from gpttui.database.base import AbstractDB, Messages, Message, MessageWithTime

class OpenAIConf(BaseModel):
    """
    Dataclass to setup OpenAI models.
    """
    timeout : int = 30
    max_retries : int = 3
    model_name : str = "gpt-3.5-turbo"
    organization : str = ""
    api_key : str = ""

class OpenAIModel(AbstractModel):
    """
    This class allows loading and interacting with any openai model through its API.
    """
    config : OpenAIConf

    def setup(self, config: OpenAIConf, session_name: str, database: AbstractDB) -> "OpenAIModel":
        """
        Initializes the model and collects credentials for OpenAI.

        Parameters
        ----------
        model_name : str
            Model name.
        session_name : str
            Session name.
        database : AbstractDB
            Database to store the messages.

        Returns
        -------
        OpenAIModel
            Instance of the model to be used as a builder.
        """
        self.config = config
        self.session_name = session_name
        self.database = database
        openai.organization = config.organization
        openai.api_key = config.api_key
        return self

    def last_messages(self) -> Messages:
        """
        Extracts the most recent messages from the database.

        Returns
        -------
        Messages
            Most recent messages.
        """
        self.database.create_session(session_name=self.session_name)
        last_msgs = self.database.get_messages(session_name=self.session_name)
        if not len(last_msgs.values):
            msg = MessageWithTime(
                    message=Message(role="system", content=self.context),
                    timestamp=int(time.time())
                    )
            self.database.add_message(msg=msg, session_name=self.session_name)
            return self.last_messages()
        return last_msgs

    async def get_answer(self, message: str) -> str:
        """
        Obtains an answer form any message.

        Parameters
        ----------
        message : str
            Input message.

        Returns
        -------
        str
            Generated response.
        """
        last_msgs = self.last_messages()
        new_msg = MessageWithTime(
                message=Message(role="user", content=message),
                timestamp=int(time.time())
                )
        self.database.add_message(msg=new_msg, session_name=self.session_name)
        last_msgs = self.last_messages()
        response = ""
        retries = 0
        while not response and retries < self.config.max_retries:
            try:
                response = await openai.ChatCompletion.acreate(
                            model = self.config.model_name,
                            messages = last_msgs.dict()["values"],
                            request_timeout = self.config.timeout
                            )
            except Timeout:
                retries += 1
        if retries == self.config.max_retries:
            raise Timeout("Maximum number of retries achieved.")
        response = str(response.choices[0].message.content) # type: ignore
        new_msg = MessageWithTime(
                message=Message(role="assistant", content=response),
                timestamp=int(time.time())
                )
        self.database.add_message(msg=new_msg, session_name=self.session_name)
        return response



"""
This file defines the required elements to use sqlite as a database.
"""
import sqlite3
from gpttui.database.base import AbstractDB, MessageWithTime, Messages, Message
from typing import Callable, List

class SqliteDB(AbstractDB):
    """
    This class represents a database based on SQLITE.

    Attributes
    ----------
    connection : sqlite3.Connection
        Connection with a sqlite database.
    """
    connection: sqlite3.Connection

    def setup(self, **kwargs: str) -> "AbstractDB":
        """
        This method allows creating a connection with the database.

        Parameters
        ----------
        database : str
            Database filename.

        Returns
        -------
        AbstractDB
            Instance of the database.
        """
        self.connection = sqlite3.connect(kwargs["database"])
        return self

    def __write_with_connection(self, f: Callable):
        """
        This method is used to handle the sqlite cursor object for write operations.

        Parameters
        ----------
        f : Callable
            Function with a query that must be executed.
        """
        cursor = self.connection.cursor()
        f(cursor)
        self.connection.commit()

    def __read_with_connection(self, f: Callable) -> List:
        """
        This method is used to handle the sqlite cursor object for read operations.

        Parameters
        ----------
        f : Callable
            Function with a query that must be executed.
        """
        cursor = self.connection.cursor()
        f(cursor)
        result = list(cursor.fetchall())
        return result

    def create_session(self, session_name: str):
        """
        Creates a session (table) in sqlite.

        Parameters
        ----------
        session_name : str
            Session name.
        """
        f = lambda cursor: cursor.execute(
                f"""
                CREATE TABLE IF NOT EXISTS {session_name}(
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    role TEXT,
                    content TEXT,
                    timestamp INT
                    );
                """
                )
        self.__write_with_connection(f)

    def delete_session(self, session_name: str):
        """
        Deletes a session (table) in sqlite.
        
        Parameters
        ----------
        session_name : str
            Session name.
        """
        f = lambda cursor: cursor.execute(
                f"""
                DROP TABLE {session_name};
                """
                )
        self.__write_with_connection(f)

    def add_message(self, msg: MessageWithTime, session_name: str):
        """
        Saves a message (row) in the database.

        Parameters
        ----------
        msg : MessageWithTime
            Message to store.
        session_name : str
            Session name.
        """
        f = lambda cursor: cursor.execute(
                f"""
                INSERT INTO {session_name} (
                    role, content, timestamp
                    )
                VALUES (?, ?, ?);
                """,
                (msg.message.role, msg.message.content, msg.timestamp)
                )
        self.__write_with_connection(f)

    def get_messages(self, session_name: str) -> Messages:
        """
        Extracts the most recent messages from the database.

        Parameters
        ----------
        session_name : str
            Session name.
        """
        f = lambda cursor: cursor.execute(
                f"""
                SELECT
                    role, content
                FROM
                    {session_name}
                ORDER BY
                    timestamp ASC
                ;
                """
                )
        result = self.__read_with_connection(f)
        messages = Messages(values=[Message(role=x[0], content=x[1]) for x in result])
        return messages

    def close(self):
        """
        Closes the connection with the database.
        """
        self.connection.close()



"""
This file defines the main configuration options for the TUI.
"""
import os
from pathlib import Path
from pydantic import BaseModel
from typing import TypeVar, Type
from gpttui.models.openai import OpenAIConf

ConfT = TypeVar("ConfT", OpenAIConf, BaseModel)

__CSS_CONFIG = """
Prompt {
    dock: bottom;
    layout: horizontal;
    padding: 0;
    margin: 0;
}

#normal-indicator {
    dock: left;
    width: 6%;
    height: 100%;
    background: #8EC07C;
    color: #3C3836;
    content-align: center middle;
    padding: 0;
    margin: 0 0 0 1;
}

#insert-indicator {
    dock: left;
    width: 6%;
    height: 100%;
    background: #458588;
    color: #3C3836;
    content-align: center middle;
    padding: 0;
    margin: 0 0 0 1;
    display: none;
}

Input {
    dock: right;
    width: 94%;
    padding: 0 0 0 1;
    margin: 0;
}

.insert-mode #normal-indicator {
    display: none;
}

.insert-mode #insert-indicator {
    display: block;
}

Message {
    layout: horizontal;
    padding: 1 0 0 0;
    margin: 0;
}

UserText {
    dock: left;
    content-align: center middle;
    background: #B8BB26;
    color: #3C3836;
    padding: 0;
    margin: 0 0 0 1;
    height: 100%;
    width: 6%;
}

UserMessage {
    dock: right;
    width: 94%;
    padding: 0 0 0 1;
    margin: 0;
}
"""

class KeyBindings(BaseModel):
    """
    Dataclass that represents the possible keybindings.
    """
    insert: str
    normal: str
    yank: str
    paste: str
    clear: str
    quit: str
    send: str
    delete: str

def config_folder() -> Path:
    """
    Setups the config folder and returns its path.

    Returns
    -------
    Path
        Configuration path.
    """
    home_path = os.environ["HOME"]
    cfg_path = Path(os.path.join(home_path, ".config/gpttui"))
    if not cfg_path.exists():
        cfg_path.mkdir()
    return cfg_path

def config_file(path: Path, type: Type[ConfT]) -> ConfT:
    """
    Setups the config file given any configuration type.

    Parameters
    ----------
    path : str
        Configuration path.
    type : Type[ConfT]
        Configuration dataclass.

    Returns
    -------
    ConfT
        Loaded configuration.
    """
    if not path.exists():
        obj = type()
        with open(path, "w") as f:
            f.write(obj.json())
    else:
        obj = type.parse_file(path)
    return obj

def css_config():
    """
    Initializes the CSS configuration file.
    """
    cfg_path = config_folder()
    filename = cfg_path / "style.css"
    if not filename.exists():
        with open(filename, "w") as f:
            f.write(__CSS_CONFIG)

def keybindings_config() -> KeyBindings:
    """
    Initializes the keybindings.
    """
    cfg_path = config_folder()
    filename = cfg_path / "keybindings.json"
    if not filename.exists():
        filename.touch()
        keybindings = KeyBindings(
                insert="i",
                normal="escape",
                yank="y",
                paste="p",
                clear="c",
                quit="q",
                send="enter",
                delete="d",
                )
        with open(filename, "w") as f:
            f.write(keybindings.json())
    else:
        keybindings = KeyBindings.parse_file(filename)
    return keybindings
